"""Context providers for enriching code indexes with business metadata.

Context providers detect ecosystem tools (dbt, Terraform, OpenAPI, etc.)
and inject business context into symbols and file summaries during indexing.
"""

from .base import ContextProvider, FileContext, discover_providers, enrich_symbols, collect_metadata

from . import dbt  # noqa: F401
from . import git_blame  # noqa: F401
from . import laravel  # noqa: F401

__all__ = [
    "ContextProvider",
    "FileContext",
    "collect_metadata",
    "discover_providers",
    "enrich_symbols",
]
